/*
  Implementation of the alternating bit protocol on top of LNP for
  LegOS/BrickOS
  Copyright (C) 2004  Bram De Wachter

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/*! \mainpage Alternating Bit Protocol
 *
 * \section contents Package Contents
 *  
 *
 * This package implements the Alternating Bit Protocal (AB-Protocol) 
 * in C with legOS/BrickOS system calls.
 * 
 * \htmlonly
  For more information on the alternating bit protocol, look 
      <a href="http://wombat.doc.ic.ac.uk/foldoc/foldoc.cgi?Alternating+bit+protocol">here</A>
      [FOLDOC]
  \endhtmlonly
 *
 * The AB-protocol makes the communciations between Lego-RCXs reliable.
 * The importance of a reliable protocol is undeniable since the standard
 * communication primitives with LegOS/BrickOS, based on LNP guarantee only
 * the integrity of packets, <B>not</B> their delivery.
 *
 * This protocol implementation does not achive miracles : two bricks that are 
 * definately out of range will never communicate (unless using a nifty
 * ad hoc protocol). The protocol will continue to try to send the message,
 * and can notify the sender when the message was received.
 *
 * This package is multi-threaded, robust, efficient and small 
 * (in that order).
 *
 * \section interface The AB-Protocols interface
 * 
 * Three functions control the AB-Protocol :
 * 1. The init_ab_protocol, which inits all data structures, 2. ab_swrite
 * for synchronous message sending (the send function returns when the message
 * is delivered) and 3. ab_awrite for asynchronous message sending (the send
 * function returns immediately).
 *
 * Look at the ab protocol interface module documentation for more details.
 *
 * We used this implementation in highly asynchronous and multithreaded 
 * environment, and took special care to avoid race-conditions, this 
 * implementation should therefore be sufficiently robust to survive 
 * the more demanding developer.
 *
 * \section license License
 *
 * This software is released under the Gnu Public License (GPL).
 * Feel free.
 *
 * \section Compiling
 *
 * To compile the library, you must have legOS/BrickOS 0.2.6 (probably 
 * works with other versions too) and the h8300 cross-compiler (which is
 * also needed to use legOS/BrickOS) installed.
 * 
 * Next, change LEGOS_ROOT in the Makefile in src/lib so it points
 * to your legOS installation directory.
 *
 * Launch <b>Make</b> to create the <b>ablib.a</b> library.
 *
 * Any legos program should then <b>#include "ab.h"</b> (located in
 * src/include)
 * and the <b>ablib.a</b> file  must be linked with the program.
 *
 * Look at the example program.
 * 
 * \section contacts Contacts/Version
 *
 * \author Bram De Wachter (<B>bdewacht</B> <I>a-t</I> <B>ulb</B>
 * <I>d-o-t</I> <B>ac</B> <I>d-o-t</I> <B>be</B>)
 * \author  Nicolas Devos
 * \date    April, 2004
 * \version 0.2
 *

 * \htmlonly
<!-- Begin Nedstat Basic code -->
<!-- Title: Reliable IR over LNP for LegOS BrickOS -->
<!-- URL: http://www.ulb.ac.be/di/ssd/bdewacht/ab/doc/ -->
<script language="JavaScript" type="text/javascript" src="http://m1.nedstatbasic.net/basic.js">
</script>
<script language="JavaScript" type="text/javascript" >
<!--
  nedstatbasic("AC1/BQ1PebKRxVMyY3Texn+wbaBA", 0);
// -->
</script>
<noscript>
<a target="_blank" href="http://www.nedstatbasic.net/stats?AC1/BQ1PebKRxVMyY3Texn+wbaBA"><img
src="http://m1.nedstatbasic.net/n?id=AC1/BQ1PebKRxVMyY3Texn+wbaBA"
border="0" width="18" height="18"
alt="Nedstat Basic - Gratis web site statistieken
Eigen homepage website teller"></a><br>
<a target="_blank" href="http://www.nedstatbasic.net/">Gratis teller</a>
</noscript>
<!-- End Nedstat Basic code -->
  \endhtmlonly
 */


/*Alternating Bit Protocol:
  1. use init_AB_protocol(portHandler) to initialize the protocol
  2. use AB_write(message, length) to send a message
*/

#include <lnp/lnp.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <semaphore.h>
#include <time.h>

/** \defgroup ab_interface Ab protocol interface
 * \brief These functions should be called to get the AB protocol up 
 * and running
 */
//@{

/// Set to 1 to get some debugging on LCD, although I do not recommend this
#define DEBUG     0

/// Defines (in ticks) the waiting time for each loop
/// in the send thread
///
/// A message to a site will AT MOST be delayed with
/// this time before it will be sent
///
/// (of course, this is an estimate, since we are not running
/// on a realtime OS, and threading may affect these values)
#define WRITE_WAIT_TIME  10 

/// Defines (in ticks) the waiting time for a retransmission
/// to a certain site which did not send any acks
/// Timeout is calculated as follows : WRITE_TIMEOUT + (random between 0 and WRITE_TIMEOUT)
#define WRITE_TIMEOUT    100

//@}


/** \defgroup ab_internals Ab protocol internals
 * Internal function for the AB protocol
 */
//@{

#define HEADER_SZ         3

/** \brief Structure used to store incoming and outgoing messages
 */
typedef struct message_struct {
  /// The pointer to the actual message contents
  unsigned char* txt;

  /// The message length (not longer than 255-DEFAULT_HEADER_SIZE)
  unsigned char len;

  /// For use with synchronous message sending, 
  /// this semaphore will be posted when the message is acknowledged
  sem_t         sem_ack;

  /// Linked list: points to the next element
  struct message_struct *next;
} msg_s ;


/** \brief Structure used to store other communicating sites
 */
typedef struct site_struct {
  /// The distant site's id, 255 different ids supported
  unsigned char id;

  /// Send and reception bit, used in the alternating bit protocol
  /// bit0 : send
  /// bit1 : recv
  unsigned char status;

  /// Used to trigger timing of non responding sites
  /// Linked to the definition of WRITE_TIMEOUT
  time_t timeout_wait;

  /// Linked list of sites
  struct site_struct *next;

  /// Pending messages (outgoing) for this site
  msg_s * head;
} site_s ;
//@}


/** \addtogroup ab_interface */
//@{

/// In your porthandler, you can use this variable to know from where the message comes
unsigned char ab_source;


/**  \brief Use this function to asynchronously send a message.

    The message is added to the list and will be sent as soon as possible. The function returns
    immediately
    \param  message The message to send
    \param  length The length of the message. For the internal housekeeping, the implementation uses
                   3 bytes in the LNP-layer message. This means that messages up to 255-3=252 bytes can be sent.
    \param  destination The destination for the message
    \return 0 on success, 1 otherwise 
*/
int ab_awrite(const unsigned char* message, unsigned char length, unsigned char destination);


/**  \brief Use this function to synchronously send a message.

    The message is added to the list and will be sent as soon as possible. The function returns
    when an acknowledgment is received for the message
    \param  message The message to send
    \param  length The length of the message. For the internal housekeeping, the implementation uses
                   3 bytes in the LNP-layer message. This means that messages up to 255-3=252 bytes can be sent.
    \param  destination The destination for the message
    \return 0 on success, 1 otherwise 
**/
int ab_swrite(const unsigned char* message, unsigned char length, unsigned char destination);

/**
   \brief Initialises everything needed for use of reliable protocol for short messages (<253 bytes)

   \param u_portHandler The function that will be called when a message is received
   \param id            The id of the current site
  \return 0 on success 1, otherwise (memory allocation problems)
*/
int init_ab_protocol_8(void (*u_portHandler) (const unsigned char *message, unsigned char length),  
                       unsigned char id);

/**
   \brief Initialises everything needed for use of reliable protocol for long messages (<65533 bytes)

   \param  u_portHandler The function that will be called when a message is received. Since large messages
   may cause insufficient memory, a message NULL will be sent to the user when such an error occurs. In
   that case the message is lost.
   \param  id            The id of the current site
   \param max_id        The highest id in the system
   \return 0 on success 1, otherwise (memory allocation problems)

   \remark Can only be used with the function ab_swrite_16 

   \remark A future release may take care of the NULL messages sent to
   the user in case of memory allocation problems, by not accepting
   the message, so that the sender must resend its message until
   enough memory is freed. For the moment, this is not the case, and
   the sender thinks that the message is handled correctely.
*/
int init_ab_protocol_16(void (*u_portHandler) (const unsigned char *message, unsigned short length),
                        unsigned char id, unsigned char max_id);

/**  \brief Use this function to synchronously send a long message.

    The message is added to the list and will be sent as soon as possible. The function returns
    when an acknowledgment is received for the message
    \param  message The message to send
    \param  length The length of the message
    \param  destination The destination for the message
    \return number of bytes sent, 0 on error

  \remark Can only be used in conjunction with init_ab_protocol_16. (In fact this function is built
  on top of init_ab_protocol_8, so you should definitely not use any other sending function)
    
*/
int ab_swrite_16(const unsigned char* message, unsigned short length, unsigned char destination);

//@}
