#include "ab.h"
/*
  Implementation of the alternating bit protocol on top of LNP for 
  LegOS/BrickOS Long messages
  Copyright (C) 2004  Bram De Wachter

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#define TRUE                        1
#define FALSE                       0

#define CHUNK_SIZE                  7 

/** \addtogroup ab_internals */
//@{

/// Pointer to the user's porthandler (for incoming messages, dispatched to the user)
/// Limited to 65536-DEFAULT_HEADER_SIZE characters 
void (*user_portHandler_16) (const unsigned char *buf, unsigned short len);


typedef struct { 
  unsigned char * cur_msg;
  unsigned short rec_len, tot_len;
  unsigned char  error, discard;
} sreceive16_s;
static sreceive16_s * rec16;
static unsigned char  max_id16;

static void porthandler_8_to_16(const unsigned char * msg, unsigned char len) {
  sreceive16_s * rec = &rec16[ab_source];
  if (ab_source>=max_id16) {
    // !! ERROR !! max_id in init_ab_16 was too small
    // Discard messages from this site
    return;
  }
  if (rec->discard) {
    // Cannot synchronize any further... Discard all messages from this
    // site. Should never happen unless LNP f*cks up
    return;
  }
  

  if (rec->cur_msg == NULL) { 
    // Receive the first chunk of data
    memcpy(&rec->tot_len,msg,2);
    rec->cur_msg = malloc (rec->tot_len);

    if (rec->cur_msg) {
      memcpy(rec->cur_msg, msg+2, len-2);
      rec->rec_len = len-2;
    } else {
      // ! OUT OF MEMORY, CAN NOT RECEIVE THE CURRENT MESSAGE 
      // Call the user with a null message to signal this
      user_portHandler_16(NULL,rec->tot_len);
      rec->error   = TRUE;
      rec->cur_msg = (unsigned char *) 1;
    }
  } else {
    // Receive next chunks of data
    if (!rec->error) memcpy(rec->cur_msg+rec->rec_len, msg, len);    
    rec->rec_len += len;
  }

  if (rec->rec_len == rec->tot_len) {
    // Handled last chunk of data
    if (!rec->error) {
      user_portHandler_16(rec->cur_msg, rec->tot_len);
      free (rec->cur_msg);
    }
    rec->error = FALSE;
    rec->cur_msg = NULL;
  } else if (rec->rec_len > rec->tot_len) {
    // ! COMMUNICATIONS ERROR !
    // Received more than expected ... do not know what to do
    // Might be handled in a later release
    // This can only be caused when LPN accepted a crippled message
    cputs ("ERAB"); 
    rec->discard = 1;
  }

}

/**
   \brief Initialises everything needed for use of reliable protocol for long messages (<65533 bytes)

   \param  u_portHandler The function that will be called when a message is received
   \param  id            The id of the current site
   \param  max_id        The maximal site id in the system
   \return 0 on success 1, otherwise (memory allocation problems)

   \remark Can only be used with the function ab_swrite_16
*/
int init_ab_protocol_16(void (*u_portHandler) (const unsigned char *message, unsigned short length),
                        unsigned char id, unsigned char max_id) {
  unsigned char i;
  user_portHandler_16 = u_portHandler;
  max_id16            = max_id+1;
  rec16               = malloc(sizeof(sreceive16_s)*max_id16);
  if (!rec16) return 1;
  for (i = 0; i <  max_id; i++) {
    rec16[i].cur_msg = NULL; 
    rec16[i].error   = FALSE;
    rec16[i].discard = FALSE;
  }
  i = init_ab_protocol_8(porthandler_8_to_16, id);
  if (i) free (rec16);
  return i;
}

/**  \brief Use this function to synchronously send a long message.

    The message is added to the list and will be sent as soon as possible. The function returns
    when an acknowledgment is received for the message
    \param  message     The message to send
    \param  length      The length of the message
    \param  destination The destination for the message
    \return number of bytes sent, 0 on error

  \remark Can only be used in conjunction with init_ab_protocol_16. (In fact this is built
  on top of init_ab_protocol_8, so you should definitely not use any other sending function)
    
*/
int ab_swrite_16(const unsigned char* message, unsigned short length, unsigned char destination) {
  unsigned char  current_chunk_sz = length+2 > CHUNK_SIZE ? CHUNK_SIZE : length+2;
  unsigned short sent_sz          = 0;
  unsigned char * first_chunk     = malloc(current_chunk_sz);
  if (!first_chunk) return 0;

  memcpy (first_chunk,&length,2);
  memcpy (first_chunk+2, message, current_chunk_sz-2);
  if (ab_swrite(first_chunk, current_chunk_sz, destination)) {
    free (first_chunk);
    return 0;
  }
  free (first_chunk);

  sent_sz += current_chunk_sz-2;

  while (sent_sz < length) {
    current_chunk_sz = length-sent_sz > CHUNK_SIZE ? CHUNK_SIZE : length-sent_sz;
    if (ab_swrite(message+sent_sz, current_chunk_sz, destination)) return sent_sz;
    sent_sz += current_chunk_sz;
  }

  return sent_sz;
}


//@}

